#ifndef __SECPWD_PLATFORM_
#define __SECPWD_PLATFORM_
/******************************************************************************/
/*                                                                            */
/*                 X r d S e c p w d P l a t f o r m. h h                     */
/*                                                                            */
/* (c) 2005 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*   Produced by Gerri Ganis for CERN                                         */
/*                                                                            */
/* This file is part of the XRootD software suite.                            */
/*                                                                            */
/* XRootD is free software: you can redistribute it and/or modify it under    */
/* the terms of the GNU Lesser General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or (at your     */
/* option) any later version.                                                 */
/*                                                                            */
/* XRootD is distributed in the hope that it will be useful, but WITHOUT      */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or      */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public       */
/* License for more details.                                                  */
/*                                                                            */
/* You should have received a copy of the GNU Lesser General Public License   */
/* along with XRootD in a file called COPYING.LESSER (LGPL license) and file  */
/* COPYING (GPL license).  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                            */
/* The copyright holder's institutional names and contributor's names may not */
/* be used to endorse or promote products derived from this software without  */
/* specific prior written permission of the institution or contributor.       */
/******************************************************************************/

//
// crypt
//
#if defined(__linux__) || defined(__solaris__)
#include <crypt.h>
#endif
#if defined(__osf__) || defined(__sgi) || defined(__APPLE__)
extern "C" char *crypt(const char *, const char *);
#endif

//
// shadow passwords
//
#include <grp.h>

#ifdef HAVE_SHADOWPW
#include <shadow.h>
#endif

#endif
